module MsgReplay exposing
    ( Config
    , sandbox, element, document, application
    , Program
    )

{-| This module helps you set up an Elm program with automatic message-replay.


# Configuration

@docs Config


# Programs

This API mirrors [elm/browser](https://package.elm-lang.org/packages/elm/browser/latest).

@docs sandbox, element, document, application

@docs Program

-}

import Browser
import Browser.Navigation
import Html exposing (Html)
import Json.Decode
import Json.Encode
import Url exposing (Url)


{-| A program with automatic message-replay enabled.
-}
type alias Program flags model msg =
    Platform.Program flags (Model model) msg


{-| This is the configuration required for automatic message-replay.

Let's say your program is initialized in HTML like this:

    <script>
        const app = Elm.Main.init({
            node: document.body,
            flags: {
                msgs : localStorage.getItem("replay-msgs")
            }
        })

        app.ports.toCache.subscribe(msgs =>
            localStorage.setItem("replay-msgs", msgs)
        )
    </script>

The configuration for that script would be like this:

    port toCache : Maybe String -> Cmd msg

    fromCache =
        .msgs

`encodeMsg` and `msgDecoder` depends on how messages are defined.

-}
type alias Config flags msg =
    { encodeMsg : msg -> Json.Encode.Value
    , msgDecoder : Json.Decode.Decoder msg
    , fromCache : flags -> Maybe String
    , toCache : String -> Cmd msg
    }


{-| Create a “sandboxed” program that cannot communicate with the outside world. More about that [here](https://package.elm-lang.org/packages/elm/browser/latest/Browser#sandbox).
-}
sandbox :
    Config flags msg
    ->
        { init : model
        , view : model -> Html msg
        , update : msg -> model -> model
        }
    -> Program flags model msg
sandbox config app =
    Browser.element
        { update =
            update (\msg model -> ( app.update msg model, Cmd.none ))
                config.toCache
                config.encodeMsg
        , subscriptions = always Sub.none
        , view = app.view << .appModel
        , init =
            \flags ->
                init app.update
                    config.msgDecoder
                    (config.fromCache flags)
                    ( app.init, Cmd.none )
        }


{-| Create an HTML element managed by Elm. More about that [here](https://package.elm-lang.org/packages/elm/browser/latest/Browser#element).
-}
element :
    Config flags msg
    ->
        { init : flags -> ( model, Cmd msg )
        , view : model -> Html msg
        , update : msg -> model -> ( model, Cmd msg )
        , subscriptions : model -> Sub msg
        }
    -> Program flags model msg
element config app =
    Browser.element
        { update = update app.update config.toCache config.encodeMsg
        , subscriptions = app.subscriptions << .appModel
        , view = app.view << .appModel
        , init =
            \flags ->
                init (replay app.update)
                    config.msgDecoder
                    (config.fromCache flags)
                    (app.init flags)
        }


{-| Create an HTML document managed by Elm. More about that [here](https://package.elm-lang.org/packages/elm/browser/latest/Browser#document).
-}
document :
    Config flags msg
    ->
        { init : flags -> ( model, Cmd msg )
        , subscriptions : model -> Sub msg
        , update : msg -> model -> ( model, Cmd msg )
        , view : model -> Browser.Document msg
        }
    -> Program flags model msg
document config app =
    Browser.document
        { update = update app.update config.toCache config.encodeMsg
        , subscriptions = app.subscriptions << .appModel
        , view = app.view << .appModel
        , init =
            \flags ->
                init (replay app.update)
                    config.msgDecoder
                    (config.fromCache flags)
                    (app.init flags)
        }


{-| Create an application that manages Url changes. More about that [here](https://package.elm-lang.org/packages/elm/browser/latest/Browser#application).
-}
application :
    Config flags msg
    ->
        { init : flags -> Url -> Browser.Navigation.Key -> ( model, Cmd msg )
        , update : msg -> model -> ( model, Cmd msg )
        , subscriptions : model -> Sub msg
        , view : model -> Browser.Document msg
        , onUrlChange : Url -> msg
        , onUrlRequest : Browser.UrlRequest -> msg
        }
    -> Program flags model msg
application config app =
    Browser.application
        { update = update app.update config.toCache config.encodeMsg
        , subscriptions = app.subscriptions << .appModel
        , view = app.view << .appModel
        , onUrlChange = app.onUrlChange
        , onUrlRequest = app.onUrlRequest
        , init =
            \flags url key ->
                init (replay app.update)
                    config.msgDecoder
                    (config.fromCache flags)
                    (app.init flags url key)
        }


type alias Model model =
    { appModel : model
    , appMsgs : List Json.Encode.Value
    }


initModel : model -> Model model
initModel appModel =
    { appModel = appModel
    , appMsgs = []
    }


init :
    (msg -> model -> model)
    -> Json.Decode.Decoder msg
    -> Maybe String
    -> ( model, Cmd msg )
    -> ( Model model, Cmd msg )
init replayUpdate msgDecoder fromCache ( appModel, appCmd ) =
    case Maybe.andThen (Result.toMaybe << Json.Decode.decodeString (decoder replayUpdate msgDecoder appModel)) fromCache of
        Just model ->
            ( model
            , Cmd.none
            )

        Nothing ->
            ( initModel appModel
            , appCmd
            )


update :
    (msg -> model -> ( model, Cmd msg ))
    -> (String -> Cmd msg)
    -> (msg -> Json.Encode.Value)
    -> msg
    -> Model model
    -> ( Model model, Cmd msg )
update updateApp toCache encodeMsg appMsg model =
    let
        ( appModel, appCmd ) =
            updateApp appMsg model.appModel

        updatedModel =
            { model
                | appModel = appModel
                , appMsgs = encodeMsg appMsg :: model.appMsgs
            }
    in
    ( updatedModel
    , Cmd.batch
        [ appCmd
        , toCache (Json.Encode.encode 0 (encode updatedModel))
        ]
    )


encode : Model model -> Json.Encode.Value
encode model =
    Json.Encode.list identity
        model.appMsgs


decoder :
    (msg -> model -> model)
    -> Json.Decode.Decoder msg
    -> model
    -> Json.Decode.Decoder (Model model)
decoder replayUpdate msgDecoder appModel =
    Json.Decode.map (decoderReplayHelper replayUpdate msgDecoder (initModel appModel))
        (Json.Decode.map List.reverse (Json.Decode.list Json.Decode.value))


decoderReplayHelper :
    (msg -> model -> model)
    -> Json.Decode.Decoder msg
    -> Model model
    -> List Json.Encode.Value
    -> Model model
decoderReplayHelper replayUpdate msgDecoder model appMsgs =
    case appMsgs of
        appMsg :: moreAppMsgs ->
            case Json.Decode.decodeValue msgDecoder appMsg of
                Ok decodedAppMsg ->
                    decoderReplayHelper replayUpdate
                        msgDecoder
                        { model
                            | appModel = replayUpdate decodedAppMsg model.appModel
                            , appMsgs = appMsg :: model.appMsgs
                        }
                        moreAppMsgs

                Err _ ->
                    model

        [] ->
            model


replay : (msg -> model -> ( model, Cmd msg )) -> msg -> model -> model
replay updateApp msg =
    Tuple.first << updateApp msg
